/* $Id: linux.c,v 1.17 1998/06/27 00:27:36 dps Exp $ */
/* Linux specific /proc reading */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_NETIPX_IPX_H
#include <netipx/ipx.h>
#endif /* HAVE_NETIPX_IPX_H */
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "cfg_linux.h"
#include "prog.h"

/* fs type support */
#ifdef STATFS_WORKS
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif /* HAVE_SYS_VFS_H */
#endif /* STATFS_WORKS */

static int log_cmdline(int pid)
{
    char filename[FNAME_LEN];	/* File name */
    char filebuf[CMD_BYTES+1];	/* bytes of command line to report */
    int all_rd,s;		/* Read all of command? */
    FILE *f;
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    struct statfs fts;
#endif

    if (snprintf(filename, FNAME_LEN, "/proc/%d/cmdline", pid)>=FNAME_LEN)
	return 0;		/* Paranoia */
    if ((f=fopen(filename,"r"))==NULL)
	return 0;

    /* This is very paranoid by we *need* paranioa here IMHO. */
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    if (fstatfs(fileno(f), &fts)!=0)
    {
	log("Unexpected failure of statfs (error %d)", errno);
    }
    if (fts.f_type!=CPS_PROC_FS_TYPE)
    {
	log("Uh. Uh. Detected hacking attempt: wrong fs type!");
    }
#endif /* HAVE_STATFS and STATFS_FST and STATFS_WORKS */

    s=fread((void *) filebuf, sizeof(char), CMD_BYTES,  f);
    if (ferror(f) || s==0)
	return 0;
    all_rd=feof(f);
    fclose(f);
    filebuf[s]='\0';
    filebuf[CMD_BYTES]='\0';	/* Ensure termination */
    log ("pid %d: command `%s' %s", pid, filebuf,
	 (all_rd) ? "" : "(truncated)" );
    return 1;
}

void print_sock(char *txt, int len, int status, const struct sockaddr *s, int l)
{
    const struct sockaddr_in *inet;
#ifdef HAVE_NETIPX_IPX_H
    const struct sockaddr_ipx *ipx;
#endif /* HAVE_NETIPX_IPX_H */
    const struct scokaddr_un *dom;
    int i, p;

    if (status!=0)
    {
	snprintf(txt, len, "[error %s]", strerror(errno));
	return;
    }
    switch(s->sa_family)
    {
    case AF_INET:
	inet=(const struct sockaddr_in *) s;
	snprintf(txt, len, "%s, port %d", inet_ntoa(inet->sin_addr),
		 ntohs(inet->sin_port));
	return;
	/* NOT REACHED */

#ifdef HAVE_NETIPX_IPX_H
    case AF_IPX:
	ipx=(const struct sockaddr_ipx *) s;
	snprintf(txt, len, "[ipx net %0hx, node %6s, port %0ux, type %d]",
		 ipx->sipx_network, ipx->sipx_node, ipx->sipx_port,
		 ipx->sipx_type);
	return;
	/* NOT REACHED */
#endif /* HAVE_NETIPX_IPX_H */

    case AF_UNIX:
	l-=offsetof(struct sockaddr, sa_data);
	snprintf(txt, len, "[socket ");
	for (i=0; i<l && i+9<len; i++)
	{
	    txt[i+8]=s->sa_data[i];
	}
	txt[i+8]='\0';
	return;
	/* NOT REACHED */

    default:
	break;
    }
    snprintf(txt, len, "[sock type %d", s->sa_family);
    l-=sizeof(s->sa_family);
    p=strlen(txt);
    for (i=0; i<l; i++, p+=3)
    {
	snprintf(txt+p, len-p, "%02x:", s->sa_data[i]);
    }
    return;
}
	
static int fd_analysis(int pid)
{
    char filename[FNAME_LEN];	/* File name */
    struct stat stbuf;
    int i, fd;
    struct sockaddr peer, local;
    char localtxt[SOCKBUF], peertxt[SOCKBUF];
    int locst, locallen, peerst, peerlen;

#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    struct statfs fts;
#endif

    for (i=0; i<MAX_FD; i++)
    {
	if (snprintf(filename, FNAME_LEN, "/proc/%d/fd/%d", pid, i)>=FNAME_LEN)
	    continue;

	if ((fd=open(filename, O_RDONLY))<0)
	    continue;
	if (fstat(fd, &stbuf)<0)
	{
	    log("pid %d: fstat failed for fd %d", pid, i);
	    continue;
	}
	if (S_ISSOCK(stbuf.st_mode))
	{
	    peerlen=sizeof(peer);
	    peerst=getpeername(fd, &peer, &peerlen);
	    print_sock(peertxt, sizeof(peertxt), peerst, &peer, peerlen);
	    locallen=sizeof(local);
	    locst=getsockname(fd, &local, &locallen);
	    print_sock(localtxt, sizeof(localtxt), locst, &local, locallen);
	    log("pid %d: fd %d is socket from %s to %s", pid, i, 
		peertxt, localtxt);
	}
	else if (S_ISREG(stbuf.st_mode))
	{
	    log("pid %d: fd %d is regular file dev %x, inode %x", pid,
		i, stbuf.st_dev, stbuf.st_ino);
	}
	else if (S_ISDIR(stbuf.st_mode))
	{
	    log("pid %d: fd %d is directory dev %x, inode %x", pid,
		i, stbuf.st_dev, stbuf.st_ino);
	}
	else if (S_ISFIFO(stbuf.st_mode))
	    log("pid %d: fd %d is fifo\n", pid, i);
	else if (S_ISCHR(stbuf.st_mode))
	    log("pid %d: fd %d is character dev %x", pid, i, stbuf.st_rdev);
	else if (S_ISBLK(stbuf.st_mode))
	    log("pid %d: fd %d is block dev %x", pid, i, stbuf.st_rdev);
	else
	    log("pid %d: fd %d is open and odd type", pid, i);
	close(fd);
    }
}

int *os_read_stat(int pid)
{
    int *res, n, i;
    char filename[FNAME_LEN];	/* Filename buffer */
    char filebuf[STAT_BYTES];	/* Contents of file */
    FILE *f;
    size_t sz;
    char *s, *t;
    int npid, pgrp;
    char state;
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    struct statfs fts;
#endif

    res=(int *) malloc(sizeof(int)*3);
    if (snprintf(filename, FNAME_LEN, "/proc/%d/stat", pid)>=FNAME_LEN)
    {
	free(res);
	return NULL;		/* Paranoia */
    }
    if ((f=fopen(filename,"r"))==NULL)
    {
	free(res);
	return NULL;
    }

    /* This is very paranoid by we *need* paranioa here IMHO. */
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    if (fstatfs(fileno(f), &fts)!=0)
    {
	log("Unexpected failure of statfs (error %d)", errno);
    }
    if (fts.f_type!=CPS_PROC_FS_TYPE)
    {
	log("Uh. Uh. Detected hacking attempt: wrong fs type!");
    }
#endif /* HAVE_STATFS and STATFS_FST and STATFS_WORKS */

    sz=fread((void *) filebuf, sizeof(char), STAT_BYTES,  f);
    if (ferror(f) || sz==0)
    {
	free(res);
	fclose(f);
	return NULL;
    }
    fclose(f);

    /* ppid, pgrp, session. We want ppid, pgrp, session in the hope that
     * one of these is the cracker's connection. Some risks are acceptable. */
    if ((s=strrchr(filebuf, '('))==NULL)
    {
	free(res);
	return NULL;
    }
    if ((t=strrchr(filebuf, ')'))==NULL || t<s)
    {
	free(res);
	return NULL;
    }
    n=sscanf(filebuf, "%d", &npid);	/* Get pid */
    if ((t+2-filebuf)>=sz)
    {
	free(res);
	return NULL;		/* Heavy paranoia */
    }
    n+=sscanf(t+2, "%c %d %d %d", &state, &(res[0]), &(res[1]),
	    &(res[2]));
    if (n<5)
    {
	free(res);
	return NULL;
    }
#ifdef DEBUG
    fprintf(stderr, "pid=%d/%d state %c, ppid=%d pgrp=%d session=%d\n",
	    pid, npid, state, res[0], res[1], res[2]);
#endif

    /* Log process information, which hopefully tells us the hole, etc */
    *t='\0';			/* Replace ) by terminator */
    log ("pid %d: name %s, pid %d, parent %d, group %d, session %d (%c)\n",
	 pid, s+1, npid, res[0], res[1], res[2], state);
    if (!log_cmdline(pid))
	log("pid %d: could not obtain command line\n", pid);
    for (i=0; i<3; i++)
    {
	if (!log_cmdline(res[i]))
	    log("pid %d: could not obtain command line\n", res[i]);
	fd_analysis(res[i]);	/* View the files the process uses. This */
				/* hopefully gets the cracker's IP numbers. */
    }
    if (npid!=pid)
    {
	log("*Very* suspect C library, asked about pid %d, read pid %d\n",
	    pid, npid);
	free(res);
	return NULL;
    }
    if (pid!=res[2])
	os_read_stat(res[2]);
    return res;
}
